import operator
from functools import reduce
from typing import Any, Dict, Iterable, List, Optional

from flatdict import FlatDict

from accession.backends import Backend, GCBackend
from accession.file import File
from accession.task import Task


class Analysis:
    """
    Parses Cromwell workflow metadata into a searchable digraph
    """

    def __init__(
        self,
        metadata,
        raw_fastqs_keys=None,
        raw_fastqs_can_have_task=False,
        auto_populate: bool = True,
        backend: Optional[Backend] = None,
    ) -> None:
        self.files: List[File] = []
        self.tasks: List[Task] = []
        self.metadata = metadata
        self.raw_fastqs_keys = raw_fastqs_keys
        self.raw_fastqs_can_have_task = raw_fastqs_can_have_task
        self.backend = backend if backend is not None else GCBackend()
        if auto_populate:
            self.tasks = self.make_tasks()

    @property
    def workflow_id(self) -> str:
        """
        Returns the Cromwell workflow id
        """
        return self.metadata.workflow_id

    def make_tasks(self) -> List[Task]:
        """
        Makes instances of Task. If the task did not succeed then will not add it to the
        call graph.
        """
        tasks = []
        for key, value in self.metadata.content["calls"].items():
            for task in value:
                if task["executionStatus"] == "Done":
                    tasks.append(self.make_task(key, task))
        for task in tasks:
            task.output_files = self.get_or_make_files(task.outputs, task)
        # Making input files after making output files avoids creating
        # a duplicate file
        for task in tasks:
            task.input_files = self.get_or_make_files(task.inputs, used_by_task=task)
        return tasks

    # Makes an instance of task with input and output GSFile instances
    def make_task(self, task_name: str, task: Dict[str, Any]) -> Task:
        new_task = Task(task_name.split(".")[1], task)
        return new_task

    def get_or_make_files(
        self,
        section: Dict[str, Any],
        task: Optional[Task] = None,
        used_by_task: Optional[Task] = None,
    ) -> List[File]:
        """
        Makes instances of File from input or output section of task. When task=None,
        file is not associated with a task. Flattens nested dict keys, so if the section
        looks like `{"foo": {"bar": "baz", "qux": [{"quux": "corge"}]}}`, first the
        section will be flattened into
        `{"foo.bar": "baz", "foo.qux": [{"quux": "corge"}]}`, note that arrays are not
        flattened. Any files anywhere inside the array will be reached later via
        `extract_files`
        """
        files = []
        flattened = FlatDict(section, delimiter=".")
        for key, value in flattened.items():
            for filename in self.extract_files(value):
                files.append(self.get_or_make_file(key, filename, task, used_by_task))
        return files

    def get_or_make_file(
        self,
        key: str,
        filename: str,
        task: Optional[Task] = None,
        used_by_task: Optional[Task] = None,
    ) -> File:
        """
        Returns a GSFile object, makes a new one if one doesn't exist
        """
        for file in self.files:
            if filename == file.filename:
                if key not in file.filekeys:
                    file.filekeys.append(key)
                if used_by_task and used_by_task not in file.used_by_tasks:
                    file.used_by_tasks.append(used_by_task)
                return file
        new_file = self.backend.make_file(key, filename, task, used_by_task)
        self.files.append(new_file)
        return new_file

    def extract_files(self, outputs):
        """
        Extracts file names from dict values
        """
        if isinstance(outputs, str) and self.backend.is_valid_uri(outputs):
            yield outputs
        elif isinstance(outputs, list):
            for item in outputs:
                yield from self.extract_files(item)
        elif isinstance(outputs, dict):
            for key, values in outputs.items():
                yield from self.extract_files(values)

    def get_tasks(self, task_name: str) -> List[Task]:
        tasks: List[Task] = []
        for task in self.tasks:
            if task_name == task.task_name:
                tasks.append(task)
        return tasks

    def get_files(self, filekey=None, filename=None):
        files = []
        if filekey:
            for file in self.files:
                if filekey in file.filekeys:
                    files.append(file)
        if filename:
            for file in self.files:
                if filename == file.filename:
                    files.append(file)
        return list(set(files))

    @property
    def raw_fastqs(self):
        fastqs = []
        raw_fastqs_keys = ("fastq", "fastqs")
        if self.raw_fastqs_keys is not None:
            raw_fastqs_keys = self.raw_fastqs_keys
        for file in self.files:
            if any([k in file.filekeys for k in raw_fastqs_keys]):
                if not self.raw_fastqs_can_have_task:
                    if file.task is not None:
                        continue
                fastqs.append(file)
        return fastqs

    def search_up(
        self,
        start_task,
        task_name,
        filekey,
        inputs=False,
        disallow_tasks: Iterable[str] = (),
    ):
        return list(
            set(self._search_up(start_task, task_name, filekey, inputs, disallow_tasks))
        )

    def search_down(self, start_task, task_name, filekey, inputs: bool = False):
        return list(set(self._search_down(start_task, task_name, filekey, inputs)))

    def _search_up(
        self,
        start_task,
        task_name,
        filekey,
        inputs: bool = False,
        disallow_tasks: Iterable[str] = (),
    ):
        """
        Search the Analysis hierarchy up for a file matching filekey. Returns a
        generator, access with next() or list() task parameter specifies the starting
        point, task_name is target task in which filekey exists.

        The disallow_tasks input is designed to avoid conflicts that can occur when
        there is a diamond dependency of several of a task's inputs on the same parent.
        This mechanism allows for halting the search up one of those branches, avoiding
        the reporting of spurious parent files. A ValueError is raised if the task name
        being searched for is also disallowed.
        """
        if task_name in disallow_tasks:
            raise ValueError(
                f"Cannot search for files in task {task_name} since this task is disallowed"
            )
        if task_name == start_task.task_name:
            if inputs is True:
                for file in start_task.input_files:
                    if filekey in file.filekeys:
                        yield file
            else:
                for file in start_task.output_files:
                    if filekey in file.filekeys:
                        yield file
        for task_item in set(
            map(
                lambda x: x.task
                if x.task is not None
                and x.task.task_name not in disallow_tasks
                and id(start_task) != id(x.task)
                else None,
                start_task.input_files,
            )
        ):
            if task_item:
                yield from self._search_up(
                    task_item, task_name, filekey, inputs, disallow_tasks=disallow_tasks
                )

    def _search_down(self, start_task, task_name, filekey, inputs: bool = False):
        """
        Search the Analysis hierarchy down for a file matching filekey. Returns
        generator object, access with next(). Task parameter specifies the starting
        point, task_name is target task in which filekey exists.

        The third argument to the reduce provides a default in the case that the task
        has no output files.
        """
        if task_name == start_task.task_name:
            if inputs is True:
                for file in start_task.input_files:
                    if filekey in file.filekeys:
                        yield file
            else:
                for file in start_task.output_files:
                    if filekey in file.filekeys:
                        yield file
        for task_item in set(  # type: ignore
            reduce(
                operator.concat,  # type: ignore
                map(lambda x: x.used_by_tasks, start_task.output_files),
                [],
            )
        ):
            if task_item:
                yield from self._search_down(task_item, task_name, filekey, inputs)



import pytest
from google.api_core.exceptions import GoogleAPICallError, RetryError

from accession.cloud_tasks import (
    AwsCredentials,
    AwsS3Object,
    CloudTasksUploadClient,
    QueueInfo,
    UploadPayload,
)
from accession.file import GSFile


@pytest.fixture
def aws_credentials():
    return AwsCredentials(
        aws_access_key_id="foo", aws_secret_access_key="bar", aws_session_token="baz"
    )


@pytest.fixture
def aws_s3_object():
    return AwsS3Object(bucket="foo", key="bar")


@pytest.fixture
def upload_payload(mocker, aws_credentials, aws_s3_object):
    blob = mocker.Mock()
    blob.name = "name"
    blob.bucket.name = "bucket"
    mocker.patch("accession.file.GSFile.blob", mocker.PropertyMock(return_value=blob))
    gs_file = GSFile(key="foo", name="gs://bucket/name")
    return UploadPayload(
        aws_credentials=aws_credentials, aws_s3_object=aws_s3_object, gcs_blob=gs_file
    )


@pytest.fixture
def cloud_tasks_upload_client(mocker):
    mocker.patch(
        "accession.cloud_tasks.CloudTasksUploadClient.client",
        new_callable=mocker.PropertyMock(),
    )
    mocker.patch(
        "accession.cloud_tasks.CloudTasksUploadClient.logger",
        new_callable=mocker.PropertyMock(),
    )
    client = CloudTasksUploadClient(
        QueueInfo(region="us-west1", name="queue"), no_log_file=True
    )
    return client


def test_aws_credentials_get_dict(aws_credentials):
    result = aws_credentials.get_dict()
    assert result == {
        "aws_access_key_id": "foo",
        "aws_secret_access_key": "bar",
        "aws_session_token": "baz",
    }


def test_aws_s3_object_get_dict(aws_s3_object):
    result = aws_s3_object.get_dict()
    assert result == {"bucket": "foo", "key": "bar"}


def test_upload_payload_get_dict(upload_payload):
    result = upload_payload.get_dict()
    assert result["aws_s3_object"]
    assert result["aws_credentials"]
    assert result["gcs_blob"]["bucket"] == "bucket"
    assert result["gcs_blob"]["name"] == "name"


def test_upload_payload_get_bytes(upload_payload):
    result = upload_payload.get_bytes()
    assert result.startswith(b'{"aws_credentials"')


def test_upload_payload_get_task_id(upload_payload):
    result = upload_payload.get_task_id()
    assert result == "127ba03ad0ee8f4fcfe64a9172507e66"


def test_cloud_tasks_upload_client_project_id(mocker, cloud_tasks_upload_client):
    mocker.patch("google.auth.default", return_value=("foo", "project-id"))
    result = cloud_tasks_upload_client.project_id
    assert result == "project-id"


def test_cloud_tasks_upload_client_project_id_google_auth_returns_none_raises(
    mocker, cloud_tasks_upload_client
):
    mocker.patch("google.auth.default", return_value=("foo", None))
    with pytest.raises(ValueError):
        _ = cloud_tasks_upload_client.project_id


def test_cloud_tasks_upload_client_get_queue_path(mocker, cloud_tasks_upload_client):
    mocker.patch("google.auth.default", return_value=("foo", "project-id"))
    cloud_tasks_upload_client.get_queue_path()
    assert cloud_tasks_upload_client.client.queue_path.called_once_with(
        "project-id", "us-west1", "queue"
    )


def test_cloud_tasks_upload_client_get_task_name(
    mocker, cloud_tasks_upload_client, upload_payload
):
    mocker.patch.object(cloud_tasks_upload_client, "get_queue_path", return_value="foo")
    mocker.patch.object(upload_payload, "get_task_id", return_value="123")
    result = cloud_tasks_upload_client._get_task_name(upload_payload)
    assert result == "foo/tasks/123"


def test_cloud_tasks_upload_client_upload(
    mocker, cloud_tasks_upload_client, upload_payload
):
    mocker.patch.object(cloud_tasks_upload_client, "_submit_task")
    cloud_tasks_upload_client.upload(upload_payload)
    assert cloud_tasks_upload_client._submit_task.called_once_with(
        "/upload", upload_payload
    )


def test_cloud_tasks_upload_client_submit_task(
    mocker, cloud_tasks_upload_client, upload_payload
):
    mocker.patch.object(
        cloud_tasks_upload_client, "get_queue_path", return_value="queue-path"
    )
    cloud_tasks_upload_client._submit_task("/endpoint", upload_payload)
    assert cloud_tasks_upload_client.client.create_task.called_once_with(
        "queue-path", upload_payload
    )
    assert cloud_tasks_upload_client.logger.info.call_args[0][2] == "/endpoint"


def test_cloud_tasks_upload_client_submit_task_value_error(
    mocker, cloud_tasks_upload_client, upload_payload
):
    mocker.patch.object(
        cloud_tasks_upload_client, "get_queue_path", return_value="queue-path"
    )
    mocker.patch.object(
        cloud_tasks_upload_client.client, "create_task", side_effect=ValueError("error")
    )
    with pytest.raises(ValueError):
        try:
            cloud_tasks_upload_client._submit_task("/endpoint", upload_payload)
        finally:
            assert cloud_tasks_upload_client.logger.exception.called_once()


def test_cloud_tasks_upload_client_submit_task_google_api_call_error(
    mocker, cloud_tasks_upload_client, upload_payload
):
    mocker.patch.object(
        cloud_tasks_upload_client, "get_queue_path", return_value="queue-path"
    )
    mocker.patch.object(
        cloud_tasks_upload_client.client,
        "create_task",
        side_effect=GoogleAPICallError(message="foo"),
    )
    with pytest.raises(GoogleAPICallError):
        try:
            cloud_tasks_upload_client._submit_task("/endpoint", upload_payload)
        finally:
            assert cloud_tasks_upload_client.logger.exception.call_args[0][1] == "foo"


def test_cloud_tasks_upload_client_submit_task_retry_error(
    mocker, cloud_tasks_upload_client, upload_payload
):
    mocker.patch.object(
        cloud_tasks_upload_client, "get_queue_path", return_value="queue-path"
    )
    mocker.patch.object(
        cloud_tasks_upload_client.client,
        "create_task",
        side_effect=RetryError(message="foo", cause=Exception("bar")),
    )
    with pytest.raises(RetryError):
        try:
            cloud_tasks_upload_client._submit_task("/endpoint", upload_payload)
        finally:
            assert cloud_tasks_upload_client.logger.exception.call_args[0][1] == "foo"



__title__ = "accession"
__version__ = "3.0.1"
__description__ = "Tool to submit genomics pipeline outputs to the ENCODE Portal"
__url__ = "https://github.com/ENCODE-DCC/accession"
__uri__ = __url__
__author__ = "Paul Sud"
__email__ = "encode-help@lists.stanford.edu"
__license__ = "MIT"
__copyright__ = "Copyright (c) 2019 ENCODE DCC"



import json
from base64 import b64encode
from collections import UserDict
from enum import Enum
from typing import Any, Dict, List, Optional, Tuple, TypeVar, Union

from encode_utils.connection import Connection

from accession.accession_steps import FileParams

T = TypeVar("T")
U = TypeVar("U")
V = TypeVar("V", bound="EncodeFile")


class EncodeGenericObject:
    def __init__(self, portal_properties: Dict[str, Any]) -> None:
        self.portal_properties = portal_properties

    @property
    def at_id(self) -> str:
        return self.portal_properties["@id"]


class EncodeCommonMetadata(UserDict):
    """
    Class to hold common metadata shared by all posted objects. Inherits from UserDict
    so we can do things like qc.update(EncodeCommonMetadata("foo", "bar"))
    """

    def __init__(self, lab: str, award: str):
        self.data = {"lab": lab, "award": award}

    @property
    def lab(self) -> str:
        return self.data["lab"]

    @property
    def award(self) -> str:
        return self.data["award"]

    @property
    def lab_pi(self) -> str:
        return self.lab.split("/labs/")[1].split("/")[0]


class EncodeFile:
    """
    A subset of file properties is made available via @property, for convenience
    """

    ASSEMBLY = "assembly"
    GENOME_ANNOTATION = "genome_annotation"

    def __init__(self, portal_file: Dict[str, Any]):
        self._portal_file = portal_file
        self.at_id = portal_file["@id"]

    def __eq__(self, other):  # noqa: E821 # type: ignore[override]
        """
        Helpful for pytest assertions. See https://github.com/python/mypy/issues/2783
        for rationale for ignoring type.
        """
        if type(self) != type(other):
            return False
        return self.at_id == other.at_id and self.portal_file == other.portal_file

    def __str__(self):
        return self.at_id

    @property
    def portal_file(self) -> Dict[str, Any]:
        return self._portal_file

    @portal_file.setter
    def portal_file(self, value) -> None:
        new_id = value["@id"]
        if new_id != self.at_id:
            raise ValueError(
                f"Cannot update file properties, expected object with an @id of {self.at_id} but received {new_id}"
            )
        self._portal_file = value

    @property
    def accession(self) -> str:
        return self.at_id.split("/")[-2]

    @property
    def output_type(self) -> str:
        return self.portal_file["output_type"]

    @property
    def status(self) -> str:
        return self.portal_file["status"]

    @property
    def dataset(self) -> str:
        return self.portal_file["dataset"]

    @property
    def md5sum(self) -> str:
        return self.portal_file["md5sum"]

    @property
    def submitted_file_name(self) -> str:
        return self.portal_file["submitted_file_name"]

    def get(
        self, key: str, default: Optional[U] = None
    ) -> Union[Optional[T], Union[T, U]]:
        """
        Type signature here is complicated. Here we want to emulate behaviour of normal
        .get on a dict. If no default is proved, then will return an object of type T
        or None, while if a default of type U is provided, then the get will either
        return an object of type T or the default.
        """
        return self.portal_file.get(key, default)

    @property
    def step_run_id(self) -> str:
        step_run = self.portal_file.get("step_run")
        if step_run is None:
            raise ValueError(f"Could not find step run for file {self.at_id}")
        if isinstance(step_run, str):
            step_run_id = step_run
        elif isinstance(step_run, dict):
            step_run_id = step_run["@id"]
        return step_run_id

    def has_qc(self, qc_type: str) -> bool:
        """
        Checks if the portal file has a qc with the given @type, e.g. StarQualityMetric
        """
        if list(
            filter(lambda x: qc_type in x["@type"], self.portal_file["quality_metrics"])
        ):
            return True
        return False

    @staticmethod
    def filter_encode_files_by_status(
        encode_files: List[V],
        forbidden_statuses: Tuple[str, ...] = ("replaced", "revoked", "deleted"),
    ) -> List[V]:
        """
        Filter out files whose statuses are not allowed. From list of EncodeFile
        instances representing encode file objects, filter out ones whose statuses are
        not allowed.

        Args:
            encode_files (list): List containing dicts representing encode file objects.
            forbidden_statuses (list): List of statuses. If file object has one of these statuses it will be filtered out.
            Statuses that encode file can have are: uploading, upload failed, in progress, released, archived, deleted, replaced,
            revoked, content error.

        Returns:
            list: List containing the EncodeFiles whose statuses are not contained in forbidden_statuses, empty list is possible.

        Raises:
            KeyError: If some of the files do not have status (this is an indication of an error on portal).
        """
        filtered_files = [
            file for file in encode_files if file.status not in forbidden_statuses
        ]
        return filtered_files

    @staticmethod
    def from_template(
        aliases: List[str],
        assembly: str,
        common_metadata: EncodeCommonMetadata,
        dataset: str,
        derived_from: List[str],
        file_params: FileParams,
        file_size: int,
        file_md5sum: str,
        step_run_id: str,
        submitted_file_name: str,
        genome_annotation: Optional[str] = None,
        extras: Optional[Dict[str, Any]] = None,
    ) -> Dict[str, Any]:
        """
        Note that extras are used to update the dictionary at the very end. This allows
        for callbacks to override properties like output_type that may have been
        specified in the template.
        """
        obj = {
            "status": "uploading",
            "aliases": aliases,
            "file_format": file_params.file_format,
            "output_type": file_params.output_type,
            "assembly": assembly,
            "dataset": dataset,
            "step_run": step_run_id,
            "derived_from": derived_from,
            "file_size": file_size,
            "md5sum": file_md5sum,
            "submitted_file_name": submitted_file_name,
        }
        if file_params.file_format_type:
            obj["file_format_type"] = file_params.file_format_type
        if genome_annotation is not None:
            obj["genome_annotation"] = genome_annotation
        if extras is not None:
            obj.update(extras)
        obj[Connection.PROFILE_KEY] = "file"
        obj.update(common_metadata)
        return obj


class EncodeAnalysis:
    """
    Class representing a nascent Analysis object. Not intended to be used to represent
    existing analyses.
    """

    PROFILE = "analysis"

    def __init__(
        self,
        files: List[EncodeFile],
        documents: List[EncodeGenericObject],
        lab_pi: str,
        workflow_id: str,
        pipeline_version: Optional[str] = None,
    ) -> None:
        """
        `documents` is a list of `EncodeGenericObject` that gives no access to the
        document internals except for the `@id`.
        """
        self.files = files
        self.aliases = self.get_aliases(lab_pi, workflow_id)
        self.documents = documents
        self.pipeline_version = pipeline_version

    def __eq__(
        self, other
    ) -> bool:  # type: ignore  # https://github.com/python/mypy/issues/2783
        """
        Helpful for pytest assertions. Should use # type: ignore[override], but flake8
        gets confused by that, raises F821.
        """
        if type(self) != type(other):
            return False
        return sorted([f.at_id for f in self.files]) == sorted(
            [f.at_id for f in other.files]
        )

    def __str__(self) -> str:
        return str([str(f) for f in self.files])

    @staticmethod
    def get_aliases(lab: str, workflow_id: str) -> List[str]:
        return [f"{lab}:{workflow_id}"]

    def get_portal_object(self) -> Dict[str, Any]:
        """
        Obtain the portal-postable dict representation of the analysis.
        """
        if self.files is None:
            raise ValueError("Cannot create payload for analysis without files")
        payload = {
            Connection.PROFILE_KEY: self.PROFILE,
            "aliases": self.aliases,
            "documents": [d.at_id for d in self.documents],
            "files": [f.at_id for f in self.files],
        }
        if self.pipeline_version is not None:
            payload["pipeline_version"] = self.pipeline_version
        return payload


class EncodeExperiment:
    INTERNAL_STATUS_KEY = "internal_status"
    INTERNAL_STATUS_POST_ACCESSIONING = "pipeline completed"
    ALLOWED_REPLICATE_STATUSES = ("released", "in progress")

    def __init__(self, portal_experiment: Dict[str, Any]):
        self.at_id = portal_experiment["@id"]
        self.portal_properties = portal_experiment

    @property
    def assay_term_name(self) -> str:
        return self.portal_properties["assay_term_name"]

    @property
    def is_replicated(self):
        return self.get_number_of_biological_replicates() > 1

    @property
    def accession(self) -> str:
        return self.at_id.split("/")[-2]

    def get_number_of_biological_replicates(self) -> int:
        bio_reps = set(
            [
                rep.get("biological_replicate_number")
                for rep in self.portal_properties["replicates"]
                if rep["status"] in self.ALLOWED_REPLICATE_STATUSES
            ]
        )
        return len([rep for rep in bio_reps if rep is not None])

    def get_number_of_technical_replicates(self) -> int:
        tech_reps = set(
            [
                rep.get("technical_replicate_number")
                for rep in self.portal_properties["replicates"]
                if rep["status"] in self.ALLOWED_REPLICATE_STATUSES
            ]
        )
        return len([rep for rep in tech_reps if rep is not None])

    def get_patchable_internal_status(self):
        return {
            self.INTERNAL_STATUS_KEY: self.INTERNAL_STATUS_POST_ACCESSIONING,
            Connection.ENCID_KEY: self.at_id,
        }

    def get_patchable_analysis_object(
        self, analysis_object_at_id: str
    ) -> Dict[str, Union[str, List[str]]]:
        return {
            "analysis_objects": [analysis_object_at_id],
            Connection.ENCID_KEY: self.at_id,
            Connection.PROFILE_KEY: "experiment",
        }


class EncodeAttachment:
    def __init__(
        self,
        contents: bytes,
        filename: str,
        mime_type: Optional[str] = None,
        additional_extension: Optional[str] = None,
    ):
        """
        Filename is not technically required, but if you don't specify the attachment
        will say "file not available" when you view it on the portal, so we require one
        here.
        """
        self.contents = contents
        self.filename = filename
        self.mime_type = mime_type
        self.additional_extension = additional_extension

    @staticmethod
    def encode_attachment_data(data: bytes) -> str:
        """
        Encodes the attachment data into a b64 datastring
        input: data as bytes object
        Output: data as string, encoded as b64
        """
        return b64encode(data).decode("utf-8")

    @staticmethod
    def get_bytes_from_dict(input_dict: Dict, encoding: str = "utf-8") -> bytes:
        """
        Useful for encoding QC JSON files into bytes for posting as attachments
        """
        return json.dumps(input_dict).encode(encoding)

    def make_download_link(self, additional_extension: str) -> str:
        return self.filename.split("/")[-1] + additional_extension

    def get_portal_object(
        self, mime_type: Optional[str] = None, additional_extension: str = ""
    ) -> Dict[str, str]:
        """
        Obtain the postable representation of the attachment. If `mime_type` or
        `extension` are specified here they will override any values specified during
        instantiation. `extension` is a string that will be appended to the download
        link to trick the mime validation code in certain cases.
        """
        if mime_type is None:
            if self.mime_type is None:
                raise ValueError("Must specify mime type for attachment via __init__")
            mime_type = self.mime_type
        if self.additional_extension is not None:
            additional_extension = self.additional_extension
        attachment_object = {
            "type": mime_type,
            "download": self.make_download_link(additional_extension),
            "href": "data:{};base64,{}".format(
                mime_type, self.encode_attachment_data(self.contents)
            ),
        }
        return attachment_object


class EncodeQualityMetric:
    def __init__(self, payload: Dict[str, Any], file_id: str):
        if not file_id:
            raise ValueError(
                "No file_id specified, QC metric needs an accessioned file"
            )
        self.files = [file_id]
        self.payload = payload

    def get_portal_object(self) -> Dict[str, Any]:
        self.payload.update({"status": "in progress", "quality_metric_of": self.files})
        return self.payload


class EncodeStepRun:
    def __init__(self, portal_step_run: Dict[str, Any]):
        self.at_id = portal_step_run["@id"]
        self.portal_step_run = portal_step_run


class EncodeDocumentType(Enum):
    WorkflowMetadata = "workflow metadata"


class EncodeDocument:
    PROFILE = "document"

    def __init__(
        self,
        attachment: EncodeAttachment,
        common_metadata: EncodeCommonMetadata,
        document_type: EncodeDocumentType,
        aliases: Optional[List[str]] = None,
    ) -> None:
        """
        Instatiates a document from the appropriate metadata. Note that document_type is
        a member of an enum, see `EncodeDocumentType` for possible variants.
        """
        self.attachment = attachment
        self.common_metadata = common_metadata
        self.document_type = document_type
        self.aliases = aliases

    def get_portal_object(self) -> Dict[str, Any]:
        attachment = self.attachment.get_portal_object()
        payload = {
            Connection.PROFILE_KEY: self.PROFILE,
            "document_type": self.document_type.value,
            "attachment": attachment,
        }
        payload.update(self.common_metadata)
        if self.aliases is not None:
            payload["aliases"] = self.aliases
        return payload



from contextlib import suppress as does_not_raise
from pathlib import Path

import pytest

from accession.analysis import Analysis
from accession.metadata import FileMetadata
from accession.task import Task


@pytest.fixture
def analysis(mock_gc_backend):
    analysis = Analysis(
        FileMetadata("foo"), backend=mock_gc_backend, auto_populate=False
    )
    return analysis


class StubMetadata:
    content = {
        "workflowRoot": "gs://foo/bar",
        "calls": {
            "wf.a": [
                {"executionStatus": "RetryableFailure"},
                {"executionStatus": "Done", "inputs": {}, "outputs": {}},
            ]
        },
    }


def test_invalid_metadata_raises():
    """
    Trying to create an Analysis from invalid JSON should raise an error.
    """
    current_dir = Path(__file__).resolve()
    metadata_json = current_dir.parent / "data" / "invalid_metadata.json"
    with pytest.raises(Exception):
        Analysis(metadata_json)


@pytest.mark.filesystem
def test_make_tasks(empty_analysis):
    assert len(empty_analysis.files) == 0
    empty_analysis.tasks = empty_analysis.make_tasks()
    assert len(empty_analysis.files) == 208
    assert len(empty_analysis.tasks) == 38


def test_make_tasks_does_not_add_failed_tasks(mocker, mock_gc_backend):
    mocker.patch("accession.analysis.Analysis.get_or_make_files")
    metadata = StubMetadata()
    analysis = Analysis(metadata, backend=mock_gc_backend)
    result = analysis.make_tasks()
    assert len(result) == 1


@pytest.mark.filesystem
def test_make_task(normal_analysis):
    inputs = {"foo": "gs://foo/foo"}
    outputs = {"baz": "gs://foo/baz"}
    docker_image = "ms-dos/latest"
    task = {"inputs": inputs, "outputs": outputs, "dockerImageUsed": docker_image}
    task_name = "atac.my_task"
    new_task = normal_analysis.make_task(task_name, task)
    assert new_task.task_name == "my_task"
    assert new_task.inputs == inputs
    assert new_task.outputs == outputs
    assert new_task.docker_image == docker_image


def test_analysis_get_or_make_files_simple_structure(mocker, analysis):
    mocker.patch.object(analysis, "get_or_make_file", return_value="bar")
    outputs = {
        "one_percent_footprints_bigbed": "gs://foo/footprints.fps.0.01.bb",
        "pe_fastqs": ["gs://foo/ENCFF123ABC.fastq.gz", "gs://foo/ENCFF456ABC.fastq.gz"],
    }
    result = analysis.get_or_make_files(outputs)
    assert len(result) == 3
    analysis.get_or_make_file.assert_any_call(
        "one_percent_footprints_bigbed", "gs://foo/footprints.fps.0.01.bb", None, None
    )
    analysis.get_or_make_file.assert_any_call(
        "pe_fastqs", "gs://foo/ENCFF123ABC.fastq.gz", None, None
    )
    analysis.get_or_make_file.assert_any_call(
        "pe_fastqs", "gs://foo/ENCFF456ABC.fastq.gz", None, None
    )


def test_analysis_get_or_make_files_flattens_nested_dicts(mocker, analysis):
    mocker.patch.object(analysis, "get_or_make_file", return_value="bar")
    outputs = {
        "analysis": {
            "one_percent_footprints_bigbed": "gs://foo/footprints.fps.0.01.bb",
            "replicate": {
                "read_length": 76,
                "number": 1,
                "se_fastqs": None,
                "pe_fastqs": [
                    {
                        "R1": "gs://foo/ENCFF123ABC.fastq.gz",
                        "R2": "gs://foo/ENCFF456ABC.fastq.gz",
                    }
                ],
                "adapters": [{"sequence_R1": "ATGC"}],
            },
        }
    }
    result = analysis.get_or_make_files(outputs)
    assert len(result) == 3
    analysis.get_or_make_file.assert_any_call(
        "analysis.one_percent_footprints_bigbed",
        "gs://foo/footprints.fps.0.01.bb",
        None,
        None,
    )
    analysis.get_or_make_file.assert_any_call(
        "analysis.replicate.pe_fastqs", "gs://foo/ENCFF123ABC.fastq.gz", None, None
    )
    analysis.get_or_make_file.assert_any_call(
        "analysis.replicate.pe_fastqs", "gs://foo/ENCFF456ABC.fastq.gz", None, None
    )


@pytest.mark.filesystem
def test_get_or_make_file_empty_analysis(empty_analysis):
    assert len(empty_analysis.files) == 0
    task = {"inputs": {}, "outputs": {}, "dockerImageUsed": ""}
    empty_task = Task("bam2ta", task)
    # Make a file when it doesn't exist
    filekey = "ta"
    filename = (
        "gs://encode-pipeline-test-runs/atac/2099edd0-0399-46ba-941c-abdcea355c1c/call-"
        "bam2ta/shard-0/glob-199637d3015dccbe277f621a18be9eb4/ENCFF599TJR.merged.nodup."
        "tn5.tagAlign.gz"
    )
    empty_analysis.get_or_make_file(filekey, filename, empty_task)
    assert len(empty_analysis.files) == 1
    assert empty_analysis.files[0].filename == filename
    # Get file if it exists, don't make a duplicate
    empty_analysis.get_or_make_file(filekey, filename, empty_task)
    assert len(empty_analysis.files) == 1


@pytest.mark.filesystem
def test_get_or_make_file_normal_analysis(normal_analysis):
    assert len(normal_analysis.files) == 208


@pytest.mark.filesystem
def test_workflow_id(normal_analysis):
    expected = "2099edd0-0399-46ba-941c-abdcea355c1c"
    assert normal_analysis.workflow_id == expected


@pytest.mark.filesystem
def test_get_files(normal_analysis):
    files = normal_analysis.get_files(filekey="ta")
    assert len(files) == 9
    assert all(["ta" in file.filekeys for file in files])
    filename = files[0].filename
    files2 = normal_analysis.get_files(filename=filename)
    assert filename == files2[0].filename
    assert len(files2) == 1


def test_get_tasks(normal_analysis):
    tasks = normal_analysis.get_tasks("bam2ta")
    tasks2 = normal_analysis.get_tasks("qc_report")
    assert len(tasks) == 2
    assert all([task.task_name == "bam2ta" for task in tasks])
    assert len(tasks2) == 1
    assert tasks2[0].task_name == "qc_report"


@pytest.mark.filesystem
def test_search_up(normal_analysis):
    task = normal_analysis.get_tasks("qc_report")[0]
    files = normal_analysis.search_up(task, "bam2ta", "ta")
    assert len(files) == 2
    assert "ta" in files[0].filekeys
    task2 = normal_analysis.get_tasks("bam2ta")[0]
    files2 = normal_analysis.search_up(task2, "trim_adapter", "fastqs", inputs=True)
    assert len(files2) == 2
    task3 = normal_analysis.get_tasks("ataqc")[0]
    files3 = normal_analysis.search_up(task3, "trim_adapter", "fastqs", inputs=True)
    # Notice how tree nodes widen the search scope
    assert len(files3) == 4


@pytest.mark.parametrize(
    "condition,disallow_tasks,expected",
    [
        (does_not_raise(), ("choose_ctl"), ["gs://foo/bar/ENCFF295YVN.nodup.bam"]),
        (
            does_not_raise(),
            (),
            [
                "gs://foo/bar/ENCFF295YVN.nodup.bam",
                "gs://foo/bar/ENCFF279HDE.nodup.bam",
            ],
        ),
        (pytest.raises(ValueError), ("bam2ta"), []),
    ],
)
def test_search_up_disallow_tasks(
    mock_metadata, mock_gc_backend, condition, disallow_tasks, expected
):
    metadata = {
        "workflowRoot": "gs://foo/bar/",
        "calls": {
            "chip.bam2ta": [
                {
                    "executionStatus": "Done",
                    "dockerImageUsed": "arch:latest",
                    "inputs": {"bam": "gs://foo/bar/ENCFF295YVN.nodup.bam"},
                    "outputs": {
                        "ta": "gs://foo/bar/call-bam2ta/shard-0/glob-35/ENCFF295YVN.nodup.tagAlign.gz"
                    },
                },
                {
                    "executionStatus": "Done",
                    "dockerImageUsed": "arch:latest",
                    "inputs": {"bam": "gs://foo/bar/ENCFF279HDE.nodup.bam"},
                    "outputs": {
                        "ta": "gs://foo/bar/call-bam2ta/shard-1/glob-35/ENCFF279HDE.nodup.tagAlign.gz"
                    },
                },
            ],
            "chip.macs2_signal_track": [
                {
                    "executionStatus": "Done",
                    "dockerImageUsed": "arch:latest",
                    "inputs": {
                        "tas": [
                            "gs://foo/bar/call-bam2ta/shard-0/glob-35/ENCFF295YVN.nodup.tagAlign.gz",
                            "gs://foo/bar/call-choose_ctl/glob-99/ctl_for_rep1.tagAlign.gz",
                        ]
                    },
                    "outputs": {
                        "fc_bw": "gs://foo/bar/call-macs2_signal_track/shard-0/glob-10/ENCFF295YVN.bigwig",
                        "pval_bw": "gs://foo/bar/call-macs2_signal_track/shard-0/glob-89/ENCFF295YVN.bigwig",
                    },
                }
            ],
            "chip.choose_ctl": [
                {
                    "executionStatus": "Done",
                    "dockerImageUsed": "arch:latest",
                    "inputs": {
                        "tas": [
                            "gs://foo/bar/call-bam2ta/shard-0/glob-35/ENCFF295YVN.nodup.tagAlign.gz",
                            "gs://foo/bar/call-bam2ta/shard-1/glob-35/ENCFF279HDE.nodup.tagAlign.gz",
                        ]
                    },
                    "outputs": {
                        "chosen_ctl_tas": [
                            "gs://foo/bar/call-choose_ctl/glob-99/ctl_for_rep1.tagAlign.gz",
                            "gs://foo/bar/call-choose_ctl/glob-99/ctl_for_rep2.tagAlign.gz",
                        ]
                    },
                }
            ],
        },
    }
    mock_metadata.content = metadata
    analysis = Analysis(mock_metadata, backend=mock_gc_backend)
    start_task = analysis.get_tasks("macs2_signal_track")[0]
    with condition:
        results = analysis.search_up(
            start_task=start_task,
            task_name="bam2ta",
            filekey="bam",
            inputs=True,
            disallow_tasks=disallow_tasks,
        )
        result = [i.filename for i in results]
        assert sorted(result) == sorted(expected)


@pytest.mark.filesystem
def test_search_down(normal_analysis):
    task = normal_analysis.get_tasks("bam2ta")[0]
    files = normal_analysis.search_down(task, "ataqc", "html")
    assert len(files) == 2
    assert "html" in files[0].filekeys
    files2 = normal_analysis.search_down(task, "qc_report", "report")
    assert len(files2) == 1
    assert "report" in files2[0].filekeys


def test_search_down_no_output_files_does_not_raise(mocker):
    """
    Regression test related to PIP-1163. If the current file being searched does not
    have any output files, then the _search_down should not throw an error in the call
    to reduce(), and instead yield nothing.
    """
    mocker.patch(
        "builtins.open", mocker.mock_open(read_data='{"workflowRoot": "gs://foo/bar"}')
    )
    analysis = Analysis(FileMetadata("foo"), auto_populate=False, backend="")
    no_outputs_task = Task("foo", {"inputs": {"bar": "gs://baz"}, "outputs": {}})
    result = list(analysis._search_down(no_outputs_task, "searching_for", "a_file"))
    assert result == []


@pytest.fixture
def empty_analysis(mock_gc_backend, metadata_json_path):  # noqa: F811
    empty = Analysis(
        FileMetadata(metadata_json_path), auto_populate=False, backend=mock_gc_backend
    )
    return empty



import json
import logging
from hashlib import md5
from typing import Dict, Optional

import attr
import google.auth
from google.api_core.exceptions import GoogleAPICallError, RetryError
from google.cloud import tasks_v2
from typing_extensions import TypedDict

from accession.file import GSFile
from accession.logger_factory import logger_factory

APP_ENGINE_UPLOAD_ENDPOINT = "/upload"

AppEngineHttpRequest = TypedDict(
    "AppEngineHttpRequest",
    {"headers": Dict[str, str], "http_method": str, "relative_uri": str, "body": bytes},
    total=False,
)
Task = TypedDict(
    "Task", {"name": str, "app_engine_http_request": AppEngineHttpRequest}, total=False
)


@attr.s(auto_attribs=True)
class QueueInfo:
    name: str
    region: str


@attr.s(auto_attribs=True)
class AwsCredentials:
    aws_access_key_id: str
    aws_secret_access_key: str
    aws_session_token: str

    def get_dict(self) -> Dict[str, str]:
        return {
            "aws_access_key_id": self.aws_access_key_id,
            "aws_secret_access_key": self.aws_secret_access_key,
            "aws_session_token": self.aws_session_token,
        }


@attr.s(auto_attribs=True)
class AwsS3Object:
    bucket: str
    key: str

    def get_dict(self) -> Dict[str, str]:
        return {"bucket": self.bucket, "key": self.key}


class UploadPayload:
    def __init__(
        self,
        aws_credentials: AwsCredentials,
        aws_s3_object: AwsS3Object,
        gcs_blob: GSFile,
    ) -> None:
        self.aws_credentials = aws_credentials
        self.aws_s3_object = aws_s3_object
        self.gcs_blob = gcs_blob

    def __str__(self) -> str:
        return json.dumps(self.get_dict())

    def get_dict(self) -> Dict[str, Dict[str, str]]:
        """
        Gets the `dict` payload to use as the request body when submitting the task,
        e.g.
        ```
        {
            "aws_credentials": {
                "aws_access_key_id": "foo",
                "aws_secret_access_key": "bar",
                "aws_session_token": "baz",
            },
            "aws_s3_object": {"bucket": "s3", "key": "object"},
            "gcs_blob": {"bucket": "cool", "name": "object"},
        }
        ```
        """
        return {
            "aws_credentials": self.aws_credentials.get_dict(),
            "aws_s3_object": self.aws_s3_object.get_dict(),
            "gcs_blob": {
                "bucket": self.gcs_blob.blob.bucket.name,
                "name": self.gcs_blob.blob.name,
            },
        }

    def get_bytes(self) -> bytes:
        """
        Return the encoded, JSON-serialized representation of the object
        """
        return json.dumps(self.get_dict()).encode()

    def get_task_id(self) -> str:
        """
        Returns the md5 hash of the payload as a serialized JSON string. This is useful
        for task deduplication on the queue.
        """
        return md5(self.get_bytes()).hexdigest()


class CloudTasksUploadClient:
    """
    Helper class for submitting upload tasks to Google Cloud Tasks
    """

    def __init__(
        self, queue_info: QueueInfo, log_file_path="accession.log", no_log_file=False
    ) -> None:
        """
        `location` and `queue` refer to the region and name of the queue, respectively.
        """
        self.queue_info = queue_info
        self._client = None
        self._project_id: Optional[str] = None
        self._logger: Optional[logging.Logger] = None
        self.log_file_path = log_file_path
        self.no_log_file = no_log_file

    @property
    def logger(self) -> logging.Logger:
        """
        Creates the instance's logger if it doesn't already exist, then returns the
        logger instance. Configured to log both to stderr (StreamHandler default) and to
        a log file.
        """
        if self._logger is None:
            logger = logger_factory(__name__, self.log_file_path, self.no_log_file)
            self._logger = logger
        return self._logger

    @property
    def project_id(self) -> str:
        """
        Raises `ValueError` if `google.auth.default()` cannot determine the project.
        """
        if self._project_id is None:
            _, project_id = google.auth.default()
            if project_id is None:
                raise ValueError(
                    (
                        "Could not determine project ID, try setting with `gcloud "
                        "config set project [PROJECT ID]`"
                    )
                )
            self._project_id = project_id
        return self._project_id

    @property
    def client(self) -> tasks_v2.CloudTasksClient:
        """
        Create a Cloud Tasks client, see
        https://googleapis.dev/python/cloudtasks/latest/gapic/v2/api.html
        """
        if self._client is None:
            self._client = tasks_v2.CloudTasksClient()
        return self._client

    def get_queue_path(self) -> str:
        """
        Return the fully qualified path to the queue.
        """
        return self.client.queue_path(
            project=self.project_id,
            location=self.queue_info.region,
            queue=self.queue_info.name,
        )

    def _get_task_name(self, payload: UploadPayload) -> str:
        """
        Return the properly formatted name of the task using the task's id. This isn't
        a method of UploadPayload because it's not a part of the request body that is
        sent to the upload endpoint and because knowledge of the queue info is required.
        """
        return f"{self.get_queue_path()}/tasks/{payload.get_task_id()}"

    def upload(self, payload: UploadPayload) -> None:
        """
        Wrapper to submit the payload to the upload endpoint, assumed to be `/upload`
        """
        self._submit_task(APP_ENGINE_UPLOAD_ENDPOINT, payload)

    def _submit_task(self, task_endpoint: str, payload: UploadPayload) -> None:
        """
        Given a `dict` payload to submit to Cloud Tasks, creates a new Cloud Tasks
        payload, inserts the necessary fields including the encoded request body, and
        submits the task. It is asssumed here that the task is triggered via a POST
        request to an App Engine endpoint.
        """
        parent = self.get_queue_path()
        task: Task = {
            "app_engine_http_request": {
                "body": payload.get_bytes(),
                "http_method": "POST",
                "relative_uri": task_endpoint,
                "headers": {"Content-Type": "application/json"},
            },
            "name": self._get_task_name(payload),
        }

        try:
            cloud_task = self.client.create_task(parent, task)
        except ValueError:
            self.logger.exception("Could not create task due to invalid parameters")
            raise
        except GoogleAPICallError as google_api_call_error:
            self.logger.exception(
                "Could not create task due to API error, original message: %s",
                google_api_call_error.message,
            )
            raise
        except RetryError as retry_error:
            self.logger.exception(
                "Could not create task due to Google API retries exceeded, original message: %s",
                retry_error.message,
            )
            raise
        else:
            self.logger.info(
                "Successfully created task %s targeting %s",
                cloud_task.name,
                task_endpoint,
            )



import json
from abc import ABC, abstractmethod
from pathlib import Path
from typing import Any, Dict, List, Optional, TextIO, Union

import WDL
from caper.caper_labels import CaperLabels

from accession.caper_helper import CaperHelper, caper_conf_exists
from accession.encode_models import EncodeAttachment


class Metadata(ABC):
    @property
    @abstractmethod
    def content(self):
        raise NotImplementedError

    @property
    def workflow_id(self) -> str:
        return self.content["id"]

    @property
    def backend_name(self) -> str:
        return self.content["labels"][CaperLabels.KEY_CAPER_BACKEND]

    def get_filename(self, prefix: str = "") -> str:
        """
        Construct an artificial filename for the metadata JSON. We do this because it
        could be possible in the future that the metadata is not actually in a file, for
        instance if we pull them directly from Caper.
        """
        delimited_prefix = f"{prefix}_" if prefix else ""
        return f"{delimited_prefix}{self.workflow_id}_metadata.json"

    def get_as_attachment(self, filename_prefix: str = "") -> EncodeAttachment:
        """
        Get the representation of the attachment on the portal
        """
        metadata_bytes = EncodeAttachment.get_bytes_from_dict(self.content)
        attachment = EncodeAttachment(
            metadata_bytes,
            self.get_filename(prefix=filename_prefix),
            mime_type="application/json",
        )
        return attachment

    def get_parsed_workflow(self) -> WDL.Tree.Document:
        return WDL.parse_document(self.content["submittedFiles"]["workflow"])


class FileMetadata(Metadata):
    def __init__(self, metadata_filepath: Union[str, Path]) -> None:
        self._metadata_filepath = metadata_filepath
        self._content: Optional[Dict[str, Any]] = None

    @property
    def content(self) -> Dict[str, Any]:
        if self._content is None:
            with open(self._metadata_filepath) as fp:
                self._content = json.load(fp)
        return self._content


class CaperMetadata(Metadata):
    def __init__(self, workflow_id_or_label: str) -> None:
        self.workflow_id_or_label = workflow_id_or_label
        self.caper_helper = CaperHelper()
        self._content: Optional[Dict[str, Any]] = None

    @property
    def content(self) -> Dict[str, Any]:
        if self._content is None:
            metadata = self.caper_helper.metadata([self.workflow_id_or_label])
            if len(metadata) != 1:
                raise ValueError("Expected one metadata JSON to be returned")
            self._content = metadata[0]
        return self._content


def metadata_factory(path_or_caper_id: str) -> Metadata:
    """
    Generates instance of FileMetadata or CaperMetadata. First assumes the input is a
    file path, if that doesn't exist it falls back to assuming it is a Caper ID or
    label if the Caper conf file exists. Raises if the metadata could not be interpreted
    as a file and the Caper conf is not present.
    """
    metadata_path = Path(path_or_caper_id)
    if metadata_path.exists():
        return FileMetadata(metadata_path)
    if caper_conf_exists():
        return CaperMetadata(path_or_caper_id)
    raise ValueError("Could not initialize metadata")


def parse_metadata_list(metadata_list_fp: TextIO) -> List[str]:
    """
    Parsed the given list of metadata paths/Caper IDs. Caper labels cannot have spaces
    so we assume that multiple entries present on a line is an error. Empty lines are
    skipped.
    """
    parsed = []
    for i, line in enumerate(metadata_list_fp):
        split = line.strip().split()
        if len(split) > 1:
            raise ValueError(
                f"Invalid metadata list, found multiple entries in line {i + 1}"
            )
        if len(split) == 0:
            continue
        parsed.append(split[0])
    if not parsed:
        raise ValueError("Metadata list is empty")
    return parsed



import pytest

from accession.accession import AccessionWgbs
from accession.analysis import Analysis


@pytest.fixture
def mock_accession_wgbs(
    mocker,
    mock_accession_gc_backend,
    mock_metadata,
    mock_accession_steps,
    server_name,
    common_metadata,
    gsfile,
):
    mocked_accession = AccessionWgbs(
        mock_accession_steps,
        Analysis(mock_metadata, backend=mock_accession_gc_backend),
        server_name,
        common_metadata,
        no_log_file=True,
    )
    mocker.patch.object(mocked_accession.analysis, "get_files", return_value=[gsfile])
    mocker.patch.object(mocked_accession.analysis, "search_down", return_value=[gsfile])
    mocker.patch.object(mocked_accession, "queue_qc", lambda output, *_: output)
    mocker.patch.object(mocked_accession, "get_attachment")
    return mocked_accession


def test_accession_wgbs_make_gembs_alignment_qc(
    mocker, mock_accession_wgbs, gsfile, encode_file_no_qc
):
    alignment_qc = {"general_reads": 3000, "pct_general_reads": 0.5}
    average_coverage_qc = {"average_coverage": {"average_coverage": 3.5}}
    mocker.patch.object(
        gsfile, "read_json", side_effect=[alignment_qc, average_coverage_qc]
    )
    qc = mock_accession_wgbs.make_gembs_alignment_qc(encode_file_no_qc, file=gsfile)
    assert qc["average_coverage"] == 3.5
    assert qc["general_reads"] == 3000
    assert qc["pct_general_reads"] == 50.0
    assert len(mock_accession_wgbs.get_attachment.mock_calls) == 2
